/**
 * Built-in generation script
 * ==========================
 *
 * This script requires a *api.json* file in the scripts directory.
 * You can get one from Appcelerator itself with (for example):
 *
 *   $ wget http://docs.appcelerator.com/titanium/data/4.1.0/api.json
 *
 * then you can run this script with:
 *
 *   $ npm run generate-built-ins
 *
 * This will write the *easy* and *simple* built-ins in the `src/lib/built-ins`
 * directory.
 */

import { createReadStream, writeFileSync } from 'fs';
import { resolve } from 'path';
import { inspect } from 'util';
import { parse } from 'JSONStream';
import chalk from 'chalk';

// Views that need a custom implementation
const customs = words`
  Ti.UI.iOS.NavigationWindow
  Ti.UI.iOS.SplitWindow
  Ti.UI.iOS.Toolbar
  Ti.UI.iPad.Popover
  Ti.UI.ListView
  Ti.UI.ListSection
  Ti.UI.MobileWeb.NavigationGroup
  Ti.UI.PickerColumn
  Ti.UI.ScrollableView
  Ti.UI.Tab
  Ti.UI.TabGroup
  Ti.UI.TableView
  Ti.UI.Window
`;

// Views that are not going to be used as elements
const skip = words`
  Ti.UI.2DMatrix
  Ti.UI.3DMatrix
  Ti.UI.Android.ProgressIndicator
  Ti.UI.Animation
  Ti.UI.AttributedString
  Ti.UI.DashboardView
  Ti.UI.iOS.Animator
  Ti.UI.ListItem
  Ti.UI.Notification
`;

// Proxies that act as views but don’t extend from Ti.UI.View
const fake = words`
  Ti.UI.TableViewSection
  Ti.UI.View
`;

// Tagnames to be renamed
const tagnames = {
  //'android-searchview': 'android-search',
  'imageview': 'image',
  'listview': 'list',
  'scrollview': 'scroll',
  'scrollableview': 'scrollable',
  'tableview': 'table',
  //'webview': 'web',
  'ios-adview': 'ios-ad',
  'ios-coverflowview': 'ios-coverflow',
};

// Views that support the <view>text</view> form, and for which property
const textProperties = {
  'Ti.UI.Label': 'text',
  'Ti.UI.Button': 'title'
};

const buildDir = resolve(__dirname, '..', 'src', 'built-ins');

// Factories

const simpleView = (tagname, apiName) => `
// Generated by scripts/generate-built-ins

import { register } from '../ReactTitaniumBridge';

register(${inspect(tagname)}, ${inspect(apiName)}, {
  textProperty: ${inspect(textProperties[apiName])},
  factory: props => ${getFactory(apiName)}(props)
});
`;

// Script

createReadStream(resolve(__dirname, './api.json'))
  .pipe(parse('*', ({ name, type, deprecated, ...rest }) => {
    const apiName = name.replace('Titanium', 'Ti');

    if (type === 'module') return;
    if (!apiName::startsWith('Ti.UI.')) return;
    if (skip::contains(apiName)) return;
    if (deprecated) return;

    const [ shortName, platform ] = apiName.toLowerCase().split('.').slice(2).reverse();

    const tagname = getTagName(platform ? `${platform}-${shortName}` : shortName);

    if (apiName::endsWith('Style')) {
      console.log(highlight(chalk.blue)`I should do something with constants from ${apiName}`);
    }
    else if (apiName::endsWith('Behavior')) {
      console.log(highlight(chalk.blue)`I should do something with the behavior ${apiName}`);
    }
    else if (apiName::endsWith('Dialog')) {
      console.log(highlight(chalk.yellow)`I should do something with the dialog ${apiName}`)
    }
    else if ((rest.extends === 'Titanium.Proxy') && !fake::contains(apiName)) {
      console.log(highlight(chalk.magenta)`I don’t know what should I do with ${apiName}`)
    }
    else if (customs::contains(apiName)) {
      try {
        require.resolve(getFilename(tagname));
        console.log(highlight(chalk.green)`<${tagname} /> builds a ${apiName}`);
      }
      catch (e) {
        console.log(highlight(chalk.red)`There’s no way to use <${tagname} /> to build a ${apiName}`);
      }
    }
    else {
      console.log(highlight(chalk.green)`<${tagname} /> builds a ${apiName}`);
      write(tagname, simpleView(tagname, apiName));
    }
  }));

// Utils

function write(tagname, source) {
  source = source.trim() + '\n';

  writeFileSync(getFilename(tagname), source);
}

function getFilename(tagname) {
  return resolve(buildDir, tagname + '.js');
}

function getFactory(apiName) {
  const [ classname, ...rest ] = apiName.split('.').reverse();
  const namespace = rest.reverse().join('.');

  return `${namespace}.create${classname}`;
}

function getTagName(tagname) {
  return tagnames[tagname] || tagname;
}

function highlight(fn){
  return ([ first, ...rest ], ...values) =>
    first + rest.map((piece, index) =>
      `${fn(values[index])}${piece}`).join('');
}

function startsWith(start) {
  return this.slice(0, start.length) === start;
}

function endsWith(end) {
  return this.slice(-end.length) === end;
}

function contains(needle) {
  return this.indexOf(needle) >= 0;
}

function words([ src ]) {
  return src.trim().split(/\s+/);
}
